// LCD drive routine for HD44780 16x2 Display
// 2012/09/30 Eleken -> revised 18/10/07 for 40x2 display
// -> revised 22/05/19 for 16x1 display which is arranged as 8x2

//#include <inttypes.h>
//#include <avr/io.h>
//#include <util/delay.h>


#include "lcd.h"

// macros
#define sbi(port, bit)	(port |= (1<<bit))
#define cbi(port, bit)	(port &= ~(1<<bit))

/*
static void LCD_Transfer_Write4(uint8_t data){

	cbi(RS_PORT, RS_PIN); // set RS
#ifdef RW_PIN
	cbi(RW_PORT, RW_PIN); // write mode
#endif

#if 0
	uint8_t ddr_port_old = DATA_DDR & (~(0x0F << DATA_PIN));
	// higher 4 bit
	sbi(E_PORT,  E_PIN );
	DATA_DDR = ddr_port_old | (0x0F << DATA_PIN); // set io port (data bus) to output
	DATA_PORT = (DATA_PORT & DATA_PORT_MASK) | ((data>>4) << DATA_PIN);
	cbi(E_PORT,  E_PIN );
	
	DATA_DDR = ddr_port_old; // houkou register wo modosu
#else
	// higher 4 bit
	DATA_PORT = (DATA_PORT & DATA_PORT_MASK) | ((data>>4) << DATA_PIN);
	sbi(E_PORT,  E_PIN );
	cbi(E_PORT,  E_PIN );
#endif
}*/

// Initialize LCD
void LCD_Init(void)
{
//	_delay_ms(15);

	// Initializing by Instruction
	LCD_Transfer_Write(0x30, 0, 1); // Function Set
	_delay_ms(5);
	LCD_Transfer_Write(0x30, 0, 1); // Function Set
	LCD_Transfer_Write(0x30, 0, 1); // Function Set

	LCD_Transfer_Write(0x20, 0, 1); // Function Set (IF = 4bit)
	if(LCD_TYPE == LCD_16X1)
	{
		LCD_Transfer_Write(0x20, 0, 0); // Function Set (IF = 4bit, 1 Lines, 5x8 Fonts)
	}
	else
	{
		LCD_Transfer_Write(0x20 | 8, 0, 0); // Function Set (IF = 4bit, 2 Lines, 5x8 Fonts)
	}
	LCD_BusyCheck();
#if SHOW_CURSEL
	LCD_Transfer_Write(0x08 | 6, 0, 0); // Display Off
#else
	LCD_Transfer_Write(0x08 | 4, 0, 0); // Display Off
#endif
	LCD_BusyCheck();
//	LCD_Transfer_Write(0x01, 0); // Display Clear
//	LCD_BusyCheck();
	LCD_Transfer_Write(0x04 | 2, 0, 0); // Entry Mode Set (Pointer Increments, Display does not shift)
	LCD_BusyCheck();

/*
	// Reset Display with POR
	LCD_Transfer_Write4(0x20); // Function Set (IF = 4bit)
	LCD_Transfer_Write(0x20 | 8, 0); // Function Set (IF = 4bit, 2 Lines, 5x8 Fonts)
	LCD_BusyCheck();
	LCD_Transfer_Write(0x08 | 4, 0); // Display On
	LCD_BusyCheck();
	LCD_Transfer_Write(0x04 | 2, 0); // Entry Mode Set
	LCD_BusyCheck();
*/
}

// 1-byte transfer (higher4: normally set 0)
void LCD_Transfer_Write(uint8_t data, uint8_t rs, uint8_t higher4){

	(rs)? sbi(RS_PORT, RS_PIN): cbi(RS_PORT, RS_PIN); // set RS
#ifdef RW_PIN
	cbi(RW_PORT, RW_PIN); // write mode
#endif
	
#if 0
	uint8_t ddr_port_old = DATA_DDR & DATA_PORT_MASK;
	// higher 4 bit
	sbi(E_PORT,  E_PIN );
	DATA_DDR = ddr_port_old | (0x0F << DATA_PIN); // set io port (data bus) to output
	DATA_PORT = (DATA_PORT & DATA_PORT_MASK) | ((data>>4) << DATA_PIN);
	cbi(E_PORT,  E_PIN );
	
	if(higher4 == 0){
		// lower 4 bit
		sbi(E_PORT,  E_PIN );
		DATA_PORT = (DATA_PORT & DATA_PORT_MASK) | ((data&0x0F) << DATA_PIN);
		cbi(E_PORT,  E_PIN );
	}
	
	DATA_DDR = ddr_port_old; // houkou register wo modosu
#else
	// higher 4 bit
	sbi(E_PORT,  E_PIN );
	DATA_PORT = (DATA_PORT & DATA_PORT_MASK) | ((data>>4) << DATA_PIN);
	cbi(E_PORT,  E_PIN );
	
	if(higher4 == 0){
		// lower 4 bit
		sbi(E_PORT,  E_PIN );
		DATA_PORT = (DATA_PORT & DATA_PORT_MASK) | ((data&0x0F) << DATA_PIN);
		cbi(E_PORT,  E_PIN );
	}

	_delay_ms(1); // additional delay
#endif
}

#if (LCD_USE_READ_FUNCTION)
// 1-byte transfer
void LCD_Transfer_Read(uint8_t* data, uint8_t rs){
	DATA_DDR = DATA_DDR & DATA_PORT_MASK; // set io port (data bus) to input
	
	uint8_t readdata = 0;

	(rs)? sbi(RS_PORT, RS_PIN): cbi(RS_PORT, RS_PIN); // set RS
#ifdef RW_PIN
	sbi(RW_PORT, RW_PIN); // read mode
	_delay_us(1);
#endif
	
	// higher 4 bit
	sbi(E_PORT,  E_PIN );
	_delay_us(1);
	readdata = ((DATA_INPUT >> DATA_PIN) << 4) & 0xF0;
	cbi(E_PORT,  E_PIN );
	
	// lower 4 bit
	sbi(E_PORT,  E_PIN );
	_delay_us(1);
	readdata |= (DATA_INPUT >> DATA_PIN) & 0x0F;
	cbi(E_PORT,  E_PIN );
	
#ifdef RW_PIN
	cbi(RW_PORT, RW_PIN);
#endif
	*data = readdata;
}
#endif


// Polling until Busy Flag = 0
void LCD_BusyCheck(void){
#ifdef RW_PIN
	uint8_t ret;
	for(;;){
		LCD_Transfer_Read(&ret, 0);
		if(!(ret & 0x80)) break;
	}
	_delay_us(10); // additional delay
#else
//	_delay_ms(1); // additional delay
#endif
}

#if (LCD_TYPE == LCD_16X1_B)
uint8_t g_lcd_pos;
#endif

void LCD_Clear(void){
	LCD_Transfer_Write(0x01, 0, 0);
	LCD_BusyCheck();
}

// Set cursel pos (pos = row*0x40 | column; row:{0,1}, column<40)
void LCD_SetTextPos(uint8_t pos){
	LCD_Transfer_Write(0x80 | pos, 0, 0);
#if (LCD_TYPE == LCD_16X1_B)
	g_lcd_pos = pos;
//	if(pos >= 8) LCD_Transfer_Write(0xC0 | pos-8, 0, 0);
#endif
	LCD_BusyCheck();
}

void LCD_Print(const char* text, uint8_t size){
	uint8_t i;
	for(i=0; i<size; i++){
		LCD_Putc(text[i]);
	}
}
void LCD_Putc(char d){
	LCD_Transfer_Write(d, 1, 0);
#if (LCD_TYPE == LCD_16X1_B)
	g_lcd_pos++;
	if(g_lcd_pos == 8){
		LCD_SetTextPos(0x40);
	}
#endif
	LCD_BusyCheck();
}
/*
void LCD_PrintBCD(const uint16_t bcd){
	if(bcd & 0xF000){
		LCD_Putc('0' + ((bcd>>12) & 0x000F));
		LCD_Putc('0' + ((bcd>>8) & 0x000F));
		LCD_Putc('0' + ((bcd>>4) & 0x000F));
		LCD_Putc('0' + ((bcd>>0) & 0x000F));
	}else if(bcd & 0x0F00){
		LCD_Putc('0' + ((bcd>>8) & 0x000F));
		LCD_Putc('0' + ((bcd>>4) & 0x000F));
		LCD_Putc('0' + ((bcd>>0) & 0x000F));
	}else if(bcd & 0x00F0){
		LCD_Putc('0' + ((bcd>>4) & 0x000F));
		LCD_Putc('0' + ((bcd>>0) & 0x000F));
	}else{
		LCD_Putc('0' + ((bcd>>0) & 0x000F));
	}
}
*/
void LCD_SetCGRAM(uint8_t startAddr, const uint8_t* font, uint8_t size)
{
	LCD_Transfer_Write(0x40 | startAddr, 0, 0); // Set CGRAM Address
	while(size--)
	{
		LCD_Transfer_Write(*font, 1, 0);
		font++;
	}
}

