// LED Checker v3
// for ATmega8L, 8MHz
// [22/05/19] Created rev_c1815@yahoo.co.jp

#include "main.h"

volatile uint8_t g_adcPhase; // 0-1: run(channel), else: idle
volatile uint8_t g_adcConversionReq;
uint16_t g_adcResult[2];

// Timer 0: 33ms
// Timer 1: 16-bit PWM (PP)
// Timer 2: 8-bit PWM (ADC)

volatile uint8_t g_irqFlag;
volatile uint8_t g_swState;
// Timer 0 (30Hz)
ISR(TIMER0_OVF_vect)
{
	static uint8_t cnt;
	
	if(++cnt>=2) // 15Hz
	{
		g_irqFlag = 1;
		cnt = 0;
		uint8_t sw_pin = (PINB & 0b10110001) | (PIND & 0xC0) >> 4; // bit 0,2,3,4,5,7
		GetSwitchState(sw_pin, (uint8_t*)&g_swState);
	}
}

ISR(TIMER2_OVF_vect)
{
	g_adcPhase = 0;
	g_adcConversionReq = 1;
}

ISR(TIMER2_COMP_vect )
{
//	g_adcPhase = 0;
//	g_adcConversionReq = 1;
}


static void ADC_StartConversion(void)
{
	// PD5 (Debug)
	PORTD |= (1<<5);
	g_adcConversionReq = 0;
	// MUX
	ADMUX = (ADMUX & 0xF0) | 0xF; //GND
	_delay_us(200); // LPF set up time
	ADMUX = (ADMUX & 0xF0) | 0; // PC0
	ADCSRA |= 0x40; // Start
	
	// Polling & channel change
	while(ADCSRA & 0x40);
	g_adcResult[0] = ADC;
	
	ADMUX = (ADMUX & 0xF0) | 0xF; //GND
	_delay_us(1);
	ADMUX = (ADMUX & 0xF0) | 1; // PC1
	ADCSRA |= 0x40; // Start
	while(ADCSRA & 0x40);
	
	g_adcResult[1] = ADC;
	g_adcPhase = 0xFF;
	
	//asm("SLEEP");
	PORTD &= ~(1<<5);
}
// ADC Interrput
ISR(ADC_vect)
{
	g_adcResult[g_adcPhase] = ADC;
	if(g_adcPhase == 0){
		g_adcPhase = 1;
		// Start next conversion
		g_adcConversionReq = 1;
	}
	else if(g_adcPhase == 1){
		g_adcPhase = 0xFF;
	}
}


static void Init(void)
{
	DDRB  = 0b01001110; // PB0, PB4,5, PB7: SW
	PORTB = 0b11110111; // SHDN_CTL = 1 (PB6), DUTY_P = L(PB3)

	DDRC  = 0b00001100;
	PORTC = 0b00111100;
	
	DDRD  = 0b00101111;  // PD6,7: SW
	PORTD = 0b11111111;
	
	
	// Sleep
	MCUCR = 0b10000000; // Sleep enable, Idle Mode
	// Note: When ADC NR Mode, Sleep instruction invokes new ADC conversion? (from ds)
	
	// UART
	/*
	UCSRA = 0b00000010; // U2X = 1
	UCSRB = 0b00001000;
	UCSRC = 0b10000110; 
	
	UBRRH = 0;
	UBRRL = 12; // 19200 bps (19200bps x 10 x 5 = 2.6ms)
	*/
	
	// Timer 0 (8M/1024/256 = 30.5Hz)
	TCCR0 = 0b00000101; // 1/1024
	TIMSK = 0b00000001; // TOIE0 = 1
	
	// Timer 1 (PWM) OC1A/1B are used
	ICR1  = 520;
	OCR1A = 20;
	OCR1B = 20;
	TCCR1A = 0b10110010; // P-P, TOP = ICR1
	TCCR1B = 0b00011001; // PS 1/1, 8M/520 = 15.4kHz
	
	// Timer 2 (PWM) OC2 is used
	TCCR2 = 0b01011110; // Fast PWM, 8M/256/256 = 120Hz
	OCR2 = 10;
	TIMSK |= 0xC0; // TOIE2 = 1
	
	// ADC
	ADMUX  = 0b01000101; // REF = AVCC, MUX=ADC5
//	ADCSRA = 0b10001101; // ADPS = 1/32 (250 kHz), Interrupt Enable
	ADCSRA = 0b10000101; // ADPS = 1/32 (250 kHz), Interrupt Disable
}

void LCD_PrintBCD(uint16_t num)
{
	uint8_t d, zs = 0;
	d = 0;
	while(num >= 10000) {
		num -= 10000;
		d++;
		zs |= 1;
	}
	if(zs) LCD_Putc('0' + d);
	
	d = 0;
	while(num >= 1000) {
		num -= 1000;
		d++;
		zs |= 1;
	}
	if(zs) LCD_Putc('0' + d);
	
	d = 0;
	while(num >= 100) {
		num -= 100;
		d++;
		zs |= 1;
	}
	if(zs) LCD_Putc('0' + d);
	
	d = 0;
	while(num >= 10) {
		num -= 10;
		d++;
		zs |= 1;
	}
	if(zs) LCD_Putc('0' + d);
	LCD_Putc('0' + num);
}
// User Font
// 0: "3.", 1: "3V", 2: "5V", 3: "1/", 4: "12", 5: "16"
const uint8_t g_userFont[] PROGMEM = {
	0x18,0x04,0x18,0x04,0x19,0x00,0x00,0x00,
	0x18,0x04,0x18,0x04,0x19,0x05,0x02,0x00,
	0x1C,0x10,0x18,0x04,0x1D,0x05,0x02,0x00,
	0x19,0x09,0x0A,0x0A,0x04,0x04,0x08,0x00,
//	0x11,0x11,0x12,0x12,0x04,0x04,0x08,0x00,
	0x17,0x11,0x11,0x17,0x14,0x14,0x17,0x00,
	0x13,0x14,0x14,0x16,0x15,0x15,0x17,0x00
};
static void LCDUpdateCGRAM(void)
{
	LCD_Transfer_Write(0x40, 0, 0); // Set CGRAM Address
	uint8_t i;
	for(i=0; i<8*6; i++)
	{
		LCD_Transfer_Write(pgm_read_byte(&g_userFont[i]), 1, 0);
	}
}

int main(void)
{
	uint16_t g_autoShutdownTimer = 0;
	
	Init();
	
	LoadEEPSetting(&g_preference);
	ICR1 = g_preference.timer1_max;
	UpdatePWMTimer(&g_preference);
	
	LCD_Init();
	LCDUpdateCGRAM();
	LCD_Clear();
	LCD_SetTextPos(0);
//	LCD_Print("1234567890123456", 16);
	LCD_Print("(C)2022 Eleken  ", 16);
	LCD_SetTextPos(0);
	LCD_Print("LED Checker v3  ", 16);
	sei();
	_delay_ms(100);
	LCD_Clear();
	
	uint8_t signature = eeprom_read_byte((uint8_t*)0x10);
	if(signature != 0x55){
//		LCD_Print("1234567890123456", 16);
		LCD_Print("EEPROM No Data! ", 16);
		_delay_ms(100);
	}
	
	// Check SW1
	if((PINB & 0x01) == 0x00) // SW1 is pushing
	{
		if(g_swState)
		{
			uint8_t sw_state = g_swState;
			g_swState = 0;
			CalibrationRoutine(sw_state, &g_preference);
		}
		
		LCD_SetTextPos(0);
		LCD_Print("T1=", 3);
		LCD_PrintBCD(g_preference.timer1_max);
		LCD_Print("   ", 3);
		return 0;
	}
	
	uint16_t cnt = 0;
	while(1)
	{
		if(g_irqFlag)
		{
			// debug
			//LCD_SetTextPos(0);
			//LCD_PrintBCD(g_adcResult[0]);
			//LCD_Print("  ", 2);
			
			g_irqFlag = 0;
			static uint8_t localCnt;
			if(++localCnt >= 15){
				localCnt = 0;
				g_autoShutdownTimer++;
			}
			
			LCDMessageUpdate(&g_preference);
		}
		
		if(g_adcConversionReq && (g_adcPhase != 0xFF))
		{
			ADC_StartConversion();
		}
		if(g_swState)
		{
			uint8_t sw_state = g_swState;
			g_swState = 0;
			g_autoShutdownTimer = 0;
			SwitchRoutine(sw_state, &g_preference);
			
			LCDMessageUpdate(&g_preference);
		}
		// Auto shutdown
		if(g_autoShutdownTimer > SHUTDOWN_THRESHOLD)
		{
			Shutdown();
		}
		asm("SLEEP");
		
	}
	
	return 0;
}

